/*
 * rfm69_spi_driver.h
 *
 *  Created on: 20. feb. 2018
 *      Author: teig
 */

#ifndef RFM69_SPI_DRIVER_H_
#define RFM69_SPI_DRIVER_H_

// https://www.semtech.com/uploads/documents/sx1231.pdf
// Page 46:
//     The SPI interface gives access to the configuration register via a synchronous full-duplex protocol corresponding to
//     CPOL = 0 and CPHA = 0 in Motorola/Freescale nomenclature. Only the slave side is implemented.
//
// Manual above:
// Radiohead:     DataMode0
// RFM69:         SPI_MODE0
// XMOS:          SPI_MODE_0 (but the 1.1.1 and 1.1.2 in spi.pdf are wrong! See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#lib_spi)

#define SPI_MODE_0_CORRECTED SPI_MODE_1      // Or SPI_MODE_0 for two logs NO difference! (also seems to work with SPI_MODE_0 28Feb2019)
                                             // AQU=065 changing values did not help 28Feb2019 (TODO remove this text)
#define SPI_CLOCK_SPEED_KHZ  100             //  1000  1   MHz also tested CLK ok and no USE_CLOCK_BLOCK (1 MHz on RadioHead)
                                             // 12500 10.5 MHz just a mess with no USE_CLOCK_BLOCK but is fine if it's defined
#define SPI_DEASSERT_TIME    (XS1_TIMER_KHZ) // 1ms. True XMOS timer tics @ 100 MHz: 1 us is 100 and 1 ms is 100*1000

// From SEMTECH SX1231H Transceiver (Rev 1 - Oct 2011, Page 44)
//
// 5.2.1. SPI Interface
//
// The SPI interface gives access to the configuration register via a synchronous full-duplex protocol corresponding to
// CPOL = 0 and CPHA = 0 in Motorola/Freescale nomenclature. Only the slave side is implemented.
//
// Three access modes to the registers are provided:
//
//   SINGLE access: an address byte followed by a data byte is sent for a write access whereas an address byte is sent
//   and a read byte is received for the read access. The NSS pin goes low at the begin of the frame and goes high after the data byte.
//
//   BURST access: the address byte is followed by several data bytes. The address is automatically incremented internally between each
//   data byte. This mode is available for both read and write accesses. The NSS pin goes low at the beginning of the frame and stay low
//   between each byte. It goes high only after the last byte transfer.
//
//   FIFO access: if the address byte corresponds to the address of the FIFO, then succeeding data byte will address the FIFO.
//   The address is not automatically incremented but is memorized and does not need to be sent between each data
//   byte. The NSS pin goes low at the beginning of the frame and stay low between each byte. It goes high only after the last byte transfer.
//
// HOW THIS IS IMPLEMENTED HERE:
//   "NSS pin" = CS pin. See MASKOF_SPI_SLAVE0_CS. Active while low
//     SINGLE access: spiRead, spiWrite
//     BURST access:  spiBurstReadSeqRegsOrFIFO, spiBurstWriteSeqRegsOrFIFO probably only meaningful with reg=REG_AESKEY1
//     FIFO  access:  spiBurstReadSeqRegsOrFIFO, spiBurstWriteSeqRegsOrFIFO with reg=REG_FIFO

// This is the bit in the SPI address that marks it as a write
#define SPI_WRITE_MASK 0x80

typedef struct { // Leading "spi" for readability, plus..
    unsigned   spi_device_index;
    unsigned   spi_clock_speed_khz;
    spi_mode_t spi_mode;          // .. "mode" also is an RFM69 term
    unsigned   spi_deassert_time; // Basic XMOS timer 100 MHz, so 100 = 1us
} spi_params_t;


// spiRead:                     rd: [wr:01-rd:41] [wr=00>rd:90]
// spiWrite                     wr: [wr:81-rd:41] [wr:84-rd:90]
// spiBurstWriteSeqRegsOrFIFO:  WR: [wr:80-rd:54] 12
//                               0[wr:0B-rd:54]  1[wr:01-rd:0B]  2[wr:01-rd:0B]  3[wr:00-rd:0B]
//                               4[wr:01-rd:0B]  5[wr:00-rd:0B]  6[wr:00-rd:0B]  7[wr:00-rd:0B]
//                               8[wr:64-rd:0B]  9[wr:00-rd:0B] 10[wr:00-rd:0B] 11[wr:00-rd:0B]
typedef uint8_t spi_byte_t; // Register "address" (=number) with or without the write bit 0x80 like "wr:01" or "wr:81"
                            // Value read back at any write byte                               like "rd:412 or "rd:90"
                            //     includes a proper read byte                                 like "rd:90"
                            // Value written                                                   like "wr:84"
                            // Value SPI_RD_BYTE2_CMD wr as second byte during a read          like "wr=00>"
                            // In other words: JUST ANY SPI BYTE
#define SPI_RD_BYTE2_CMD 0x00
#define SPI_RD_FIFO_CMD  0x00

void
spiBeginTransaction (
        client spi_master_if i_spi,
        spi_params_t         spi_params);

void
spiEndTransaction (
        client spi_master_if i_spi,
        spi_params_t         spi_params);

spi_byte_t
spiTransfer8 ( // returns value read
        const spi_byte_t     reg, // address
        client spi_master_if i_spi,
        spi_params_t         spi_params);

spi_byte_t
spiRead ( // returns value read
        const spi_byte_t     reg, // address
        client spi_master_if i_spi,
        spi_params_t         spi_params);

spi_byte_t
spiWrite ( // returns value read back on second access
        const spi_byte_t     reg, // address
        const spi_byte_t     val,
        client spi_master_if i_spi,
        spi_params_t         spi_params);

spi_byte_t
spiBurstReadSeqRegsOrFIFO ( // returns value read from address write
        const spi_byte_t     reg, // address: REG_FIFO or probably only REG_AESKEY1
        spi_byte_t           dest [], unsigned static const len, // A pair understood by the compiler
        client spi_master_if i_spi,
        spi_params_t         spi_params);

spi_byte_t
spiBurstWriteSeqRegsOrFIFO ( // returns value read from address write
        const spi_byte_t     reg, // address: REG_FIFO or probably only REG_AESKEY1
        const spi_byte_t     src [], unsigned static const len, // A pair understood by the compiler
        client spi_master_if i_spi,
        spi_params_t         spi_params);

// From RFM69.xc
//
// spi_byte_t RFM69::SPI_READ_REG  (spi_byte_t addr)
// void    RFM69::SPI_WRITE_REG (spi_byte_t addr, spi_byte_t value)

#define SPI_READ_REG(addr)        spiRead            (addr,          i_spi, spi_params) // returns value read back on second access
#define SPI_WRITE_REG(addr,value) spiWrite           (addr, value,   i_spi, spi_params) // returns value also read back on second access
#define SPI_BEGIN_TRANSACTION()   spiBeginTransaction               (i_spi, spi_params)
#define SPI_END_TRANSACTION()     spiEndTransaction                 (i_spi, spi_params)
#define SPI_TRANSFER8(reg)        spiTransfer8       (reg,           i_spi, spi_params) // returns value read back on this access

#endif
