/*
 * rfm69_spi_driver.xc
 *
 *  Created on: 20. feb. 2018
 *      Author: teig
 */

#include <xs1.h>
#include <platform.h>  // slice
#include <timer.h>     // delay_milliseconds(200), XS1_TIMER_HZ etc
#include <stdint.h>    // spi_byte_t
#include <stdio.h>     // printf
#include <xccompat.h>  // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
#include <iso646.h>    // not etc.
#include <spi.h>

#include "rfm69_globals.h"
#include "module_xc_ptr.h" // When DEBUG_SHARED_LOG_VALUE_LIB==1
#include "rfm69_spi_driver.h"

#define DEBUG_PRINT_SPI_DRIVER    0
#define DEBUG_PRINT_SPI_TRANSFER8 0

#define debug_print(fmt, ...)   do { if((DEBUG_PRINT_SPI_DRIVER==1)    and (DEBUG_PRINT_GLOBAL_LIB==1)) printf(fmt, __VA_ARGS__); } while (0)
#define debug_print8(fmt, ...)  do { if((DEBUG_PRINT_SPI_TRANSFER8==1) and (DEBUG_PRINT_GLOBAL_LIB==1)) printf(fmt, __VA_ARGS__); } while (0)

// From "spi.h"
//     "The data will be transmitted least-significant bit first."
// This is WRONG! MSB _is_ output first!
// See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#lib_spi and
// http://www.xcore.com/viewtopic.php?f=26&t=6452&p=32240#p32240

// From SEMTECH SX1231 Transceiver (Rev.7, June 2013, Page 46)
//
//     MOSI is generated by the master on the falling edge of SCK and is sampled by the slave (i.e. this SPI interface)
//     on the rising edge of SCK. MISO is generated by the slave on the falling edge of SCK.
//
//     A transfer always starts by the NSS (=CS) pin going low. MISO is high impedance when NSS is high.
//
//     The first byte is the address byte. It is made of:
//
//     * wnr bit, which is 1 for write access and 0 for read access
//     * 7 bits of address, MSB first
//
//     The second byte is a data byte, either sent on MOSI by the master in case of a write access, or received by
//     the master on MISO in case of read access. The data byte is transmitted MSB first.

#define LSB_FIRST 0 // As we saw, MSB first is correct. Observe that MY_DEVICE_TYPE as 0x24 is symmetric: 00100100,
                    // so it's a good idea to verify others. I do this by reading default values RF_BITRATEMSB_4800 (0x1A)
                    // and RF_BITRATELSB_4800 (0x0B), causing ERROR_BIT_BITRATE_DEFAULT if not ok

#if (LSB_FIRST == 1) // TODO remove
    #error LSB first is wrong!
    spi_byte_t swap_8bits (const spi_byte_t bits) {
        spi_byte_t newbits = 0;
        newbits or_eq ((bits bitand 0x01) << 7);
        newbits or_eq ((bits bitand 0x02) << 5);
        newbits or_eq ((bits bitand 0x04) << 3);
        newbits or_eq ((bits bitand 0x08) << 1);
        newbits or_eq ((bits bitand 0x10) >> 1);
        newbits or_eq ((bits bitand 0x20) >> 3);
        newbits or_eq ((bits bitand 0x40) >> 5);
        newbits or_eq ((bits bitand 0x80) >> 7);
        return newbits;
    }
    #define ARRANGE_8BITS(bits) swap_8bits(bits) // LSB first does not work, just for test when in dispair
#else
    #define ARRANGE_8BITS(bits) bits // MSB first is the correct
#endif

#ifndef DEBUG_SHARED_LOG_VALUE_LIB
    #error DEBUG_SHARED_LOG_VALUE_LIB not defined
#endif

void
spiBeginTransaction (
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    debug_print ("%u_tr_begin\n", spi_params.spi_device_index);
    i_spi.begin_transaction (spi_params.spi_device_index, spi_params.spi_clock_speed_khz, spi_params.spi_mode);
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,31);
}

void
spiEndTransaction (
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    debug_print ("%u_tr_end\n", spi_params.spi_device_index);
    i_spi.end_transaction (spi_params.spi_deassert_time);
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,30);
}

spi_byte_t
spiTransfer8 ( // returns value read
        const spi_byte_t     reg, // address
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    spi_byte_t val_read;

    val_read = i_spi.transfer8 (ARRANGE_8BITS(reg)); // Send the address with the write mask off

    debug_print8 ("%u_tr: [wr:%02X-rd:%02X]\n", spi_params.spi_device_index, ARRANGE_8BITS(reg), /* 00> */ ARRANGE_8BITS(val_read));

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,29);

    return ARRANGE_8BITS(val_read);
}

spi_byte_t
spiRead ( // returns value read
        const spi_byte_t     reg, // address
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    spi_byte_t data;
    spi_byte_t val_read;
    spi_byte_t val_write = reg bitand (compl SPI_WRITE_MASK);

    i_spi.begin_transaction (spi_params.spi_device_index, spi_params.spi_clock_speed_khz, spi_params.spi_mode);
    data = i_spi.transfer8 (ARRANGE_8BITS(val_write)); // Send the address with the write mask off
    val_read = i_spi.transfer8 (0); // reg value is read
    i_spi.end_transaction (spi_params.spi_deassert_time);

    debug_print ("%u_rd: [wr:%02X-rd:%02X] [wr=00>rd:%02X]\n", spi_params.spi_device_index, val_write, ARRANGE_8BITS(data), /* 00> */ ARRANGE_8BITS(val_read));

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,28);

    return ARRANGE_8BITS(val_read);
}

spi_byte_t
spiWrite ( // returns value read back on second access
        const spi_byte_t     reg, // address
        const spi_byte_t     val,
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    spi_byte_t data;
    spi_byte_t val_write = reg bitor SPI_WRITE_MASK;
    spi_byte_t val_read; // Only used for printing

    i_spi.begin_transaction (spi_params.spi_device_index, spi_params.spi_clock_speed_khz, spi_params.spi_mode);
    data = i_spi.transfer8 (ARRANGE_8BITS(val_write)); // Send the address with the write mask on. Ignore data value (it's zero I think)
    val_read = i_spi.transfer8 (ARRANGE_8BITS(val)); // Ignoring return (=read) value
    i_spi.end_transaction (spi_params.spi_deassert_time);

    debug_print ("%u_wr: [wr:%02X-rd:%02X] [wr:%02X-rd:%02X]\n", spi_params.spi_device_index, val_write, ARRANGE_8BITS(data), val, ARRANGE_8BITS(val_read));

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,27);

    return ARRANGE_8BITS(val_read);
}

spi_byte_t
spiBurstReadSeqRegsOrFIFO ( // returns value read from address write
        const spi_byte_t     reg, // address: REG_FIFO or probably only REG_AESKEY1
        spi_byte_t           dest [], unsigned static const len, // A pair understood by the compiler
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    spi_byte_t data = 0;
    unsigned   iof_dest = 0;
    spi_byte_t val_write = reg bitand (compl SPI_WRITE_MASK);
    spi_byte_t val_read;
    bool       eol_just_printed = false; // Just for printing, to avoid empty lines

    i_spi.begin_transaction (spi_params.spi_device_index, spi_params.spi_clock_speed_khz, spi_params.spi_mode);
    data = i_spi.transfer8 (ARRANGE_8BITS(val_write)); // Send the address with the write mask off
    debug_print ("%u_RD: [wr:%02X-rd:%02X] %d\n", spi_params.spi_device_index, ARRANGE_8BITS(val_write), ARRANGE_8BITS(data), len);

    while (iof_dest < len) {
        val_read = i_spi.transfer8 (0);
        debug_print8 ("%2u[wr:00>rd:%02X] ", iof_dest, /* 00> */ ARRANGE_8BITS(val_read));
        dest[iof_dest] = ARRANGE_8BITS(val_read);
        iof_dest++;
        eol_just_printed = ((iof_dest % 4) == 0);
        if (eol_just_printed) {
            debug_print8("%s", "\n");
        } else {}
    }
    if (not eol_just_printed) {
        debug_print8("%s", "\n");
    } else {}

    i_spi.end_transaction (spi_params.spi_deassert_time);

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,26);

    return ARRANGE_8BITS(data);
}

spi_byte_t
spiBurstWriteSeqRegsOrFIFO ( // returns value read from address write
        const spi_byte_t     reg, // address: REG_FIFO or probably only REG_AESKEY1
        const spi_byte_t     src [], unsigned static const len, // A pair understood by the compiler
        client spi_master_if i_spi,
        spi_params_t         spi_params)
{
    spi_byte_t data = 0;
    unsigned   iof_src = 0;
    spi_byte_t val_write = reg bitor SPI_WRITE_MASK;
    spi_byte_t val_read; // Just for printing
    bool       eol_just_printed = false; // Just for printing, to avoid empty lines

    i_spi.begin_transaction (spi_params.spi_device_index, spi_params.spi_clock_speed_khz, spi_params.spi_mode);
    data = i_spi.transfer8 (ARRANGE_8BITS(val_write)); // Send the address with the write mask on
    debug_print ("%u_WR: [wr:%02X-rd:%02X] %d\n", spi_params.spi_device_index, ARRANGE_8BITS(val_write), ARRANGE_8BITS(data), len);

    while (iof_src < len) {
        val_write = src[iof_src];
        val_read = i_spi.transfer8 (ARRANGE_8BITS(val_write)); // Ignoring return (=read) value
        debug_print8 ("%2u[wr:%02X-rd:%02X] ", iof_src, ARRANGE_8BITS(val_write), ARRANGE_8BITS(val_read));
        iof_src++;
        eol_just_printed = ((iof_src % 4) == 0);
        if (eol_just_printed) {
            debug_print8("%s", "\n");
        } else {}
    }
    if (not eol_just_printed) {
        debug_print8("%s", "\n");
    } else {}

    i_spi.end_transaction (spi_params.spi_deassert_time);
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,25);

    return ARRANGE_8BITS(data);
}
